# -*- coding: utf-8 -*-
"""Top-level package for autosig."""
from .autosig import Signature, autosig, param, Retval

__all__ = ["Signature", "autosig", "param", "Retval"]
__author__ = """Antonio Piccolboni"""
__email__ = "autosig@piccolboni.info"
__version__ = "__version__ = '0.10.0'"
